print("import: 'suds'")
import suds

print("import: 'suds.bindings'")
import suds.bindings

print("import: 'suds.mx'")
import suds.mx

print("import: 'suds.sax'")
import suds.sax

print("import: 'suds.transport'")
import suds.transport

print("import: 'suds.umx'")
import suds.umx

print("import: 'suds.xsd'")
import suds.xsd

print("import: 'tests'")
import tests

print("import: 'tests.external'")
import tests.external

